@@ -28,6 +28,7 @@ mod types;
 mod chainlink_price_service;
 
 use config::{DextradesConfig, ConfigBuilder};
+use crate::config::NetworkOverrides;
 use service::DextradesService;
 use stream::*;
 use std::sync::Arc;
@@ -118,6 +119,12 @@ impl PyConfigBuilder {
         Ok(())
     }
 
+    /// Set shard count for getLogs splitting (0 = derive from providers_to_race)
+    fn shard_count(&mut self, count: usize) -> PyResult<()> {
+        self.inner = self.inner.clone().shard_count(count);
+        Ok(())
+    }
+
     /// Enable or disable sharding of log requests
     fn shard_logs(&mut self, enable: bool) -> PyResult<()> {
         self.inner = self.inner.clone().shard_logs(enable);
@@ -136,6 +143,8 @@ impl PyConfigBuilder {
         self.inner = self.inner.clone().provider_strategy(strategy_enum);
         Ok(())
     }
+
+    // Note: network_overrides are currently only configurable via Client(..., network_overrides={...})
     
     /// Build the configuration and create a client
     fn build_client(&self) -> PyResult<Client> {
@@ -153,6 +162,9 @@ impl PyConfigBuilder {
         Ok(Client {
             inner: Arc::new(service),
             rpc_urls,
+            default_ordered: false,
+            default_reorder_window_blocks: None,
+            default_max_reorder_delay_ms: None,
         })
     }
 }
@@ -162,13 +174,17 @@ impl PyConfigBuilder {
 pub struct Client {
     inner: Arc<DextradesService>,
     rpc_urls: Vec<String>,
+    // Defaults for ordering behavior when stream_swaps() omits explicit args
+    default_ordered: bool,
+    default_reorder_window_blocks: Option<u64>,
+    default_max_reorder_delay_ms: Option<u64>,
 }
 
 #[pymethods]
 impl Client {
     /// Create a new client with the given RPC URLs and configuration
     #[new]
-    #[pyo3(signature = (rpc_urls, max_concurrent_requests=10, cache_size=1000, providers_to_race=2, shard_logs=false, provider_strategy="race"))]
+    #[pyo3(signature = (rpc_urls, max_concurrent_requests=10, cache_size=1000, providers_to_race=2, shard_logs=false, provider_strategy="race", network_overrides=None, default_ordered=false, default_reorder_window_blocks=None, default_max_reorder_delay_ms=None))]
     fn new(
         rpc_urls: Vec<String>,
         max_concurrent_requests: usize,
@@ -176,6 +192,10 @@ impl Client {
         providers_to_race: usize,
         shard_logs: bool,
         provider_strategy: &str,
+        network_overrides: Option<PyObject>,
+        default_ordered: bool,
+        default_reorder_window_blocks: Option<u64>,
+        default_max_reorder_delay_ms: Option<u64>,
     ) -> PyResult<Self> {
         init_logger(); // Initialize the logger
         let rt = pyo3_async_runtimes::tokio::get_runtime();
@@ -189,24 +209,33 @@ impl Client {
             )),
         };
 
+        // Build config with optional overrides
+        let mut cfg = DextradesConfig { default_rpc_urls: rpc_urls_clone, max_concurrent_requests, cache_size, providers_to_race, shard_logs, provider_strategy: provider_strategy_enum, ..Default::default() };
+        if let Some(nw) = network_overrides {
+            Python::with_gil(|py| -> PyResult<()> {
+                let any = nw.bind(py);
+                if let Ok(dict) = any.downcast::<pyo3::types::PyDict>() {
+                    let mut ov = NetworkOverrides::default();
+                    if let Ok(Some(val)) = dict.get_item("native_wrapped") { if let Ok(s) = val.extract::<String>() { ov.native_wrapped = Some(s); } }
+                    if let Ok(Some(val)) = dict.get_item("native_usd_aggregator") { if let Ok(s) = val.extract::<String>() { ov.native_usd_aggregator = Some(s); } }
+                    if let Ok(Some(val)) = dict.get_item("stable_addresses") { if let Ok(v) = val.extract::<Vec<String>>() { ov.stable_addresses = v; } }
+                    if let Ok(Some(val)) = dict.get_item("warmup_tokens") { if let Ok(v) = val.extract::<Vec<String>>() { ov.warmup_tokens = v; } }
+                    cfg.network_overrides = Some(ov);
+                }
+                Ok(())
+            })?;
+        }
+
         let service = rt
-            .block_on(async {
-                DextradesService::new(DextradesConfig {
-                    default_rpc_urls: rpc_urls_clone,
-                    max_concurrent_requests,
-                    cache_size,
-                    providers_to_race,
-                    shard_logs,
-                    provider_strategy: provider_strategy_enum,
-                    ..Default::default()
-                })
-                .await
-            })
+            .block_on(async { DextradesService::new(cfg).await })
             .map_err(|e: error::DextradesError| -> PyErr { e.to_py_err() })?;
 
         Ok(Self {
             inner: Arc::new(service),
             rpc_urls,
+            default_ordered,
+            default_reorder_window_blocks,
+            default_max_reorder_delay_ms,
         })
     }
 
@@ -321,7 +350,7 @@ impl Client {
     ///     enrich_timestamps: Whether to enrich with block timestamps (slower)
     ///     max_concurrent_chunks: Maximum number of chunks to process concurrently (default: 3)
     ///     batches: Return Arrow batches if True, individual swaps if False (default: False)
-    #[pyo3(signature = (protocols, from_block, to_block, address=None, batch_size=None, enrich_timestamps=None, enrich_usd=None, max_concurrent_chunks=None, batches=false))]
+    #[pyo3(signature = (protocols, from_block, to_block, address=None, batch_size=None, enrich_timestamps=None, enrich_usd=None, max_concurrent_chunks=None, routers=None, ordered=None, reorder_window_blocks=None, max_reorder_delay_ms=None, batches=false, order_mode=None, allowed_lateness_blocks=None, watermark_timeout_ms=None))]
     fn stream_swaps(
         &self,
         py: Python<'_>,
@@ -333,7 +362,14 @@ impl Client {
         enrich_timestamps: Option<bool>,
         enrich_usd: Option<bool>,
         max_concurrent_chunks: Option<usize>,
+        routers: Option<Vec<String>>,
+        ordered: Option<bool>,
+        reorder_window_blocks: Option<u64>,
+        max_reorder_delay_ms: Option<u64>,
         batches: bool,
+        order_mode: Option<String>,
+        allowed_lateness_blocks: Option<u64>,
+        watermark_timeout_ms: Option<u64>,
     ) -> PyResult<PyObject> {
         // Convert protocols to Vec<String>
         let protocols_vec: Vec<String> = if let Ok(protocol_str) = protocols.extract::<String>(py) {
@@ -350,6 +386,11 @@ impl Client {
 
         if batches {
             // Return Arrow batches for performance
+            // Apply client defaults when call-site args are None
+            let ordered = ordered.unwrap_or(self.default_ordered);
+            let reorder_window_blocks = reorder_window_blocks.or(self.default_reorder_window_blocks);
+            let max_reorder_delay_ms = max_reorder_delay_ms.or(self.default_max_reorder_delay_ms);
+
             let arrow_stream = stream_dex_swaps(
                 py,
                 protocols_vec,
@@ -361,10 +402,22 @@ impl Client {
                 enrich_timestamps,
                 enrich_usd,
                 max_concurrent_chunks,
+                routers.clone(),
+                Some(ordered),
+                reorder_window_blocks,
+                max_reorder_delay_ms,
+                order_mode.clone(),
+                allowed_lateness_blocks,
+                watermark_timeout_ms,
             )?;
             Ok(Py::new(py, arrow_stream)?.into())
         } else {
             // Return individual swaps (default)
+            // Apply client defaults when call-site args are None
+            let ordered = ordered.unwrap_or(self.default_ordered);
+            let reorder_window_blocks = reorder_window_blocks.or(self.default_reorder_window_blocks);
+            let max_reorder_delay_ms = max_reorder_delay_ms.or(self.default_max_reorder_delay_ms);
+
             let arrow_stream = stream_dex_swaps(
                 py,
                 protocols_vec,
@@ -376,6 +429,13 @@ impl Client {
                 enrich_timestamps,
                 enrich_usd,
                 max_concurrent_chunks,
+                routers.clone(),
+                Some(ordered),
+                reorder_window_blocks,
+                max_reorder_delay_ms,
+                order_mode,
+                allowed_lateness_blocks,
+                watermark_timeout_ms,
             )?;
             let swap_stream = DextradesSwapStream::new(arrow_stream);
             Ok(Py::new(py, swap_stream)?.into())
@@ -394,7 +454,7 @@ impl Client {
     ///     to_block: Ending block number
     ///     address: Pool address to filter by (None for all pools)
     ///     enrich_timestamps: Whether to enrich with block timestamps (slower)
-    #[pyo3(signature = (protocols, from_block, to_block, address=None, enrich_timestamps=None, enrich_usd=None))]
+    #[pyo3(signature = (protocols, from_block, to_block, address=None, enrich_timestamps=None, enrich_usd=None, routers=None))]
     fn stream_individual_swaps(
         &self,
         py: Python<'_>,
@@ -404,6 +464,7 @@ impl Client {
         address: Option<String>,
         enrich_timestamps: Option<bool>,
         enrich_usd: Option<bool>,
+        routers: Option<Vec<String>>,
     ) -> PyResult<PyObject> {
         // Convert protocols to Vec<String>
         let protocols_vec: Vec<String> = if let Ok(protocol_str) = protocols.extract::<String>(py) {
@@ -428,6 +489,7 @@ impl Client {
             self.rpc_urls.clone(),
             enrich_timestamps,
             enrich_usd,
+            routers,
         )?;
         
         // Return individual swaps
