@@ -157,10 +157,9 @@ impl RpcOrchestrator {
         let provider_limits = RpcProviderLimits::from_url(rpc_url);
         info!("Configuring rate limiting for {}: {} CU/s", rpc_url, provider_limits.compute_units_per_second);
 
-        // Create the provider with HTTP connection and call batching
+        // Create the provider with HTTP connection, batching and retry backoff
         let provider_builder = ProviderBuilder::new();
         
-        // Create provider with call batching layer
         let provider = provider_builder
             .layer(CallBatchLayer::new().wait(config.batch_delay))
             .connect_http(url);
@@ -269,42 +268,67 @@ impl RpcOrchestrator {
         retry_policy.execute_with_retry(|| {
             let request_factory = &request_factory;
             async move {
-                // Get all healthy providers for true racing
+                // Hedged racing: start with one provider; add more after a short delay until success
                 let race_count = self
                     .config
                     .providers_to_race
                     .max(1)
                     .min(self.providers.len());
-                let providers_to_race = self.get_healthy_providers(race_count);
+                let mut providers_to_race = self.get_healthy_providers(race_count);
 
                 if providers_to_race.is_empty() {
                     return Err(eyre::eyre!("No healthy providers available for racing"));
                 }
 
-                info!("🏁 Racing {} across {} providers simultaneously", operation_name, providers_to_race.len());
+                // Dynamic hedge delay based on recent p95, with sane bounds
+                let learned_ms = self.global_rpc_metrics.get_op_p95_latency_ms(operation_name).await;
+                let hedge_delay = learned_ms
+                    .map(|ms| std::time::Duration::from_millis(ms.saturating_div(2).clamp(50, 400)))
+                    .unwrap_or_else(|| std::time::Duration::from_millis(150));
                 let start_time = std::time::Instant::now();
-                
-                // Create concurrent racing tasks
-                let racing_futures: Vec<_> = providers_to_race.into_iter()
-                    .map(|provider| {
-                        let request_future = request_factory(provider.clone());
-                        tokio::spawn(async move {
-                            request_future.await
-                        })
-                    })
-                    .collect();
-                
-                // Race all requests - first successful response wins
-                match futures::future::select_ok(racing_futures).await {
-                    Ok((result, _remaining)) => {
-                        let duration = start_time.elapsed();
-                        info!("✅ Racing {} won in {:?}", operation_name, duration);
-                        result
-                    }
-                    Err(e) => {
-                        let duration = start_time.elapsed();
-                        warn!("❌ All providers failed for {} after {:?}: {}", operation_name, duration, e);
-                        Err(eyre::eyre!("All racing providers failed for {}: {}", operation_name, e))
+                info!("🏁 Hedged racing {} with up to {} providers", operation_name, providers_to_race.len());
+
+                // Launch the first request immediately
+                let mut tasks = futures::stream::FuturesUnordered::new();
+                let first = providers_to_race.remove(0);
+                tasks.push(tokio::spawn({
+                    let fut = request_factory(first.clone());
+                    async move { (first, fut.await) }
+                }));
+
+                // Gradually add hedged requests after hedge_delay until success or exhaustion
+                loop {
+                    tokio::select! {
+                        Some(joined) = tasks.next() => {
+                            match joined {
+                                Ok((provider, Ok(res))) => {
+                                    let duration = start_time.elapsed();
+                                    info!("✅ Hedged race {} won by {} in {:?}", operation_name, provider.url, duration);
+                                    self.global_rpc_metrics.record_op_latency(operation_name, duration).await;
+                                    return Ok(res);
+                                }
+                                Ok((provider, Err(e))) => {
+                                    warn!("❌ {} failed from {}: {}", operation_name, provider.url, e);
+                                    if tasks.is_empty() && providers_to_race.is_empty() {
+                                        return Err(eyre::eyre!("All racing providers failed for {}: {}", operation_name, e));
+                                    }
+                                    // continue loop to await others or add hedges
+                                }
+                                Err(e) => {
+                                    if tasks.is_empty() && providers_to_race.is_empty() {
+                                        return Err(eyre::eyre!("Join error for {}: {}", operation_name, e));
+                                    }
+                                }
+                            }
+                        }
+                        _ = tokio::time::sleep(hedge_delay), if !providers_to_race.is_empty() => {
+                            // Launch next hedge
+                            let next = providers_to_race.remove(0);
+                            tasks.push(tokio::spawn({
+                                let fut = request_factory(next.clone());
+                                async move { (next, fut.await) }
+                            }));
+                        }
                     }
                 }
             }
@@ -403,7 +427,7 @@ impl RpcOrchestrator {
             .providers_to_race
             .max(1)
             .min(self.providers.len());
-        let candidates = self.get_healthy_providers(race_count);
+        let mut candidates = self.get_healthy_providers(race_count);
 
         if candidates.is_empty() {
             self.global_rpc_metrics.record_completion();
@@ -411,52 +435,69 @@ impl RpcOrchestrator {
         }
 
         let filter_clone = filter.clone();
+        // Dynamic hedge delay for get_logs based on recent p95
+        let learned_ms = self.global_rpc_metrics.get_op_p95_latency_ms("get_logs").await;
+        let hedge_delay = learned_ms
+            .map(|ms| std::time::Duration::from_millis(ms.saturating_div(2).clamp(50, 400)))
+            .unwrap_or_else(|| std::time::Duration::from_millis(150));
         let mut tasks = futures::stream::FuturesUnordered::new();
-        for provider in candidates.into_iter() {
-            let provider_cloned = provider.clone();
-            let filter = filter_clone.clone();
-            tasks.push(tokio::spawn(async move {
-                // Execute request
-                let res = provider_cloned.provider.get_logs(&filter).await;
-                (provider_cloned, res)
-            }));
-        }
+        let start_time = std::time::Instant::now();
+        let first = candidates.remove(0);
+        let first_filter = filter_clone.clone();
+        tasks.push(tokio::spawn(async move {
+            let res = first.provider.get_logs(&first_filter).await;
+            (first, res)
+        }));
 
-        // Wait for first success, track restrictions
         let mut last_err: Option<eyre::Error> = None;
-        while let Some(joined) = tasks.next().await {
-            match joined {
-                Ok((provider, Ok(logs))) => {
-                    provider.circuit_breaker.record_success().await;
-                    provider.health_score.fetch_add(1, Ordering::Relaxed);
-                    debug!("✅ get_logs: {} logs from {}", logs.len(), provider.url);
-                    self.global_rpc_metrics.record_completion();
-                    return Ok(logs);
-                }
-                Ok((provider, Err(e))) => {
-                    provider.circuit_breaker.record_failure().await;
-                    provider.health_score.fetch_sub(5, Ordering::Relaxed);
-                    let es = e.to_string();
-                    if es.contains("32701") || es.contains("Please specify an address") {
-                        debug!("⚠️ Provider {} requires address for get_logs", provider.url);
-                        provider
-                            .requires_address_for_logs
-                            .store(true, Ordering::Relaxed);
-                    } else {
-                        debug!("❌ get_logs failed from {}: {}", provider.url, es);
+        loop {
+            tokio::select! {
+                Some(joined) = tasks.next() => {
+                    match joined {
+                        Ok((provider, Ok(logs))) => {
+                            provider.circuit_breaker.record_success().await;
+                            provider.health_score.fetch_add(1, Ordering::Relaxed);
+                            debug!("✅ get_logs: {} logs from {}", logs.len(), provider.url);
+                            self.global_rpc_metrics.record_completion();
+                            let elapsed = start_time.elapsed();
+                            self.global_rpc_metrics.record_op_latency("get_logs", elapsed).await;
+                            return Ok(logs);
+                        }
+                        Ok((provider, Err(e))) => {
+                            provider.circuit_breaker.record_failure().await;
+                            provider.health_score.fetch_sub(5, Ordering::Relaxed);
+                            let es = e.to_string();
+                            if es.contains("32701") || es.contains("Please specify an address") {
+                                debug!("⚠️ Provider {} requires address for get_logs", provider.url);
+                                provider.requires_address_for_logs.store(true, Ordering::Relaxed);
+                            } else {
+                                debug!("❌ get_logs failed from {}: {}", provider.url, es);
+                            }
+                            last_err = Some(eyre::eyre!("get_logs failed from {}: {}", provider.url, es));
+                            if tasks.is_empty() && candidates.is_empty() {
+                                self.global_rpc_metrics.record_completion();
+                                return Err(last_err.unwrap());
+                            }
+                        }
+                        Err(join_err) => {
+                            last_err = Some(eyre::eyre!("get_logs join error: {}", join_err));
+                            if tasks.is_empty() && candidates.is_empty() {
+                                self.global_rpc_metrics.record_completion();
+                                return Err(last_err.unwrap());
+                            }
+                        }
                     }
-                    last_err = Some(eyre::eyre!("get_logs failed from {}: {}", provider.url, es));
-                    continue;
                 }
-                Err(join_err) => {
-                    last_err = Some(eyre::eyre!("get_logs join error: {}", join_err));
-                    continue;
+                _ = tokio::time::sleep(hedge_delay), if !candidates.is_empty() => {
+                    let next = candidates.remove(0);
+                    let f = filter_clone.clone();
+                    tasks.push(tokio::spawn(async move {
+                        let res = next.provider.get_logs(&f).await;
+                        (next, res)
+                    }));
                 }
             }
         }
-
-        self.global_rpc_metrics.record_completion();
-        Err(last_err.unwrap_or_else(|| eyre::eyre!("All providers failed for get_logs")))
     }
 
     /// Shard get_logs by splitting the block range into subranges and issuing parallel requests
@@ -569,27 +610,46 @@ impl RpcOrchestrator {
             let provider_clone = provider.clone();
             let filter_clone = filter.clone();
             tasks.push(tokio::spawn(async move {
-                Self::execute_with_circuit_breaker(&provider_clone, 
-                    provider_clone.provider.get_logs(&filter_clone), "get_logs_assigned").await
+                let res = Self::execute_with_circuit_breaker(&provider_clone, 
+                    provider_clone.provider.get_logs(&filter_clone), "get_logs_assigned").await;
+                (start, end, res)
             }));
         }
 
         let mut all: Vec<Log> = Vec::new();
+        let mut failed_ranges: Vec<(u64,u64)> = Vec::new();
         while let Some(res) = tasks.next().await {
             match res {
-                Ok(Ok(mut logs)) => {
-                    debug!("📥 Shard completed with {} logs", logs.len());
+                Ok((start, end, Ok(mut logs))) => {
+                    debug!("📥 Shard {}-{} completed with {} logs", start, end, logs.len());
                     all.append(&mut logs);
                 }
-                Ok(Err(e)) => {
-                    warn!("⚠️ Shard failed: {}, continuing with other shards", e);
-                    // Don't fail the entire request if one shard fails
-                    // This provides graceful degradation
+                Ok((start, end, Err(e))) => {
+                    warn!("⚠️ Shard {}-{} failed: {}, will retry via hedged get_logs", start, end, e);
+                    failed_ranges.push((start, end));
                 }
                 Err(e) => return Err(eyre::eyre!("join error: {}", e)),
             }
         }
-        
+
+        // Retry failed shards via hedged get_logs
+        for (start, end) in failed_ranges {
+            let mut filter = Filter::new()
+                .from_block(start)
+                .to_block(end)
+                .event_signature(event_signature);
+            if let Some(addr) = address { filter = filter.address(addr); }
+            match self.get_logs(&filter).await {
+                Ok(mut logs) => {
+                    debug!("🔁 Fallback shard {}-{} recovered {} logs", start, end, logs.len());
+                    all.append(&mut logs);
+                }
+                Err(e) => {
+                    warn!("❌ Fallback shard {}-{} failed: {}", start, end, e);
+                }
+            }
+        }
+
         info!("✅ Provider-assigned shards completed: {} total logs", all.len());
         Ok(all)
     }
