@@ -426,6 +426,15 @@ pub fn stream_dex_swaps(
     enrich_timestamps: Option<bool>, // Optional timestamp enrichment
     enrich_usd: Option<bool>,        // Optional USD pricing enrichment
     max_concurrent_chunks: Option<usize>, // NEW: Control chunk-level concurrency
+    routers: Option<Vec<String>>,    // Optional router whitelist (tx.to addresses)
+    // Ordering controls (original)
+    ordered: Option<bool>,
+    reorder_window_blocks: Option<u64>,
+    max_reorder_delay_ms: Option<u64>,
+    // Streaming-native aliases/presets
+    order_mode: Option<String>,                  // "immediate" | "balanced" | "strict"
+    allowed_lateness_blocks: Option<u64>,        // alias of reorder_window_blocks
+    watermark_timeout_ms: Option<u64>,           // alias of max_reorder_delay_ms
 ) -> PyResult<DextradesArrowStream> {
     if rpc_urls.is_empty() {
         return Err(pyo3::exceptions::PyValueError::new_err(
@@ -450,7 +459,7 @@ pub fn stream_dex_swaps(
     let enrich_timestamps = enrich_timestamps.unwrap_or(false);
 
     // Build configuration synchronously
-    let config = if let Some(batch_size_param) = batch_size {
+    let mut config = if let Some(batch_size_param) = batch_size {
         DextradesConfig::builder()
             .rpc_urls(rpc_urls.clone())
             .streaming_batch_size(batch_size_param)
@@ -464,6 +473,7 @@ pub fn stream_dex_swaps(
         }
         cfg
     };
+    if let Some(r) = routers { config.router_whitelist = Some(r); }
 
     // Create service using the runtime (blocking only for init)
     let service = match rt.block_on(async { DextradesService::new(config).await }) {
@@ -501,10 +511,52 @@ pub fn stream_dex_swaps(
         Arc::new(base)
     };
 
+    // Get effective values from service config
+    let batch_size = service.config().effective_batch_size();
+    let max_concurrent_chunks = service.config().effective_max_concurrent_batches();
+
     // Event processing pipeline for deduplication and ordering
+    // Resolve order mode and alias parameters
+    let mode = order_mode.as_deref().map(|s| s.to_ascii_lowercase()).unwrap_or_default();
+    let order_enabled_from_mode = match mode.as_str() {
+        "immediate" => Some(false),
+        "balanced" | "strict" => Some(true),
+        _ => None,
+    };
+    let order_enabled = ordered.or(order_enabled_from_mode).unwrap_or(false);
+
+    // Resolve lateness/window
+    let window_from_alias = allowed_lateness_blocks;
+    let window_from_mode = match mode.as_str() {
+        "immediate" => Some(0u64),
+        "balanced" => None, // derive dynamically below
+        "strict" => Some(64u64),
+        _ => None,
+    };
+    let window_explicit = reorder_window_blocks.or(window_from_alias).or(window_from_mode);
+    let reorder_blocks = if order_enabled {
+        if let Some(w) = window_explicit {
+            w
+        } else {
+            // Derive a robust default from concurrency and batch size; allow wide enough buffer
+            let mc = max_concurrent_chunks as u64;
+            std::cmp::max(32, batch_size * mc * 2)
+        }
+    } else { 0 };
+
+    // Resolve watermark timeout
+    let timeout_from_alias = watermark_timeout_ms;
+    let timeout_from_mode = match mode.as_str() {
+        "immediate" => Some(100u64),
+        "balanced" => Some(500u64),
+        "strict" => Some(1500u64),
+        _ => None,
+    };
+    let reorder_delay_ms = max_reorder_delay_ms.or(timeout_from_alias).or(timeout_from_mode).unwrap_or_else(|| if order_enabled { 500 } else { 100 });
+    let reorder_delay = std::time::Duration::from_millis(reorder_delay_ms);
     let event_pipeline_config = PipelineConfig {
-        reorder_window_blocks: 0,
-        max_reorder_delay: std::time::Duration::from_millis(100),
+        reorder_window_blocks: reorder_blocks,
+        max_reorder_delay: reorder_delay,
         pruning_interval: std::time::Duration::from_secs(300),
     };
     let event_pipeline = match rt.block_on(async { EventPipeline::new(event_pipeline_config, Arc::clone(&service)).await }) {
@@ -517,10 +569,6 @@ pub fn stream_dex_swaps(
         }
     };
 
-    // Get effective values from service config
-    let batch_size = service.config().effective_batch_size();
-    let max_concurrent_chunks = service.config().effective_max_concurrent_batches();
-
     // Start async task to process the stream using pre-built context
     let service_for_task = Arc::clone(&service);
     let pipeline_for_task = Arc::clone(&pipeline);
@@ -620,88 +668,79 @@ pub fn stream_dex_swaps(
 
                     let mut all_extracted_events = Vec::new();
 
-                    // Process each protocol
-                    for protocol in &protocols {
-                        // Get event signature and extract function for this protocol
-                        let (event_signature, extract_fn): (alloy::primitives::B256, fn(&[alloy::rpc::types::Log]) -> Vec<SwapEvent>) = match protocol.as_str() {
-                            "uniswap_v2" => (crate::protocols::uniswap_v2::get_swap_event_signature(), crate::protocols::uniswap_v2::extract_swaps),
-                            "uniswap_v3" => (crate::protocols::uniswap_v3::get_swap_event_signature(), crate::protocols::uniswap_v3::extract_swaps),
-                            _ => {
-                                eprintln!("[Stream Task] Unknown protocol: {}", protocol);
-                                continue;
-                            }
-                        };
-
-                        // Fetch logs using the RPC service
-                        let mut filter = alloy::rpc::types::Filter::new()
-                            .from_block(batch_start)
-                            .to_block(batch_end)
-                            .event_signature(event_signature);
-
-                        // Add pool filter if specified
-                        if let Some(pool_addr) = pool_filter {
-                            filter = filter.address(pool_addr);
-                        }
-
-                        let logs = if service.config().shard_logs {
-                            let result = match service.config().provider_strategy {
-                                crate::config::ProviderStrategy::Race => {
-                                    // Use racing shards (existing behavior)
-                                    service.rpc_service().get_logs_sharded(
-                                        event_signature,
-                                        pool_filter,
-                                        batch_start,
-                                        batch_end,
-                                        service.config().providers_to_race,
-                                    ).await
-                                }
-                                crate::config::ProviderStrategy::Shard => {
-                                    // Use provider-assigned shards (new linear scaling)
-                                    service.rpc_service().get_logs_provider_assigned_shards(
-                                        event_signature,
-                                        pool_filter,
-                                        batch_start,
-                                        batch_end,
-                                        service.config().providers_to_race,
-                                    ).await
+                    // Process each protocol concurrently for this chunk
+                    let mut proto_tasks = FuturesUnordered::new();
+                    for protocol in protocols.clone() {
+                        let service = Arc::clone(&service);
+                        let protocol_name = protocol.clone();
+                        let pool_filter_local = pool_filter;
+                        proto_tasks.push(Box::pin(async move {
+                            // Get event signature + extractor for this protocol
+                            let (event_signature, extract_fn): (alloy::primitives::B256, fn(&[alloy::rpc::types::Log]) -> Vec<SwapEvent>) = match protocol_name.as_str() {
+                                "uniswap_v2" => (crate::protocols::uniswap_v2::get_swap_event_signature(), crate::protocols::uniswap_v2::extract_swaps),
+                                "uniswap_v3" => (crate::protocols::uniswap_v3::get_swap_event_signature(), crate::protocols::uniswap_v3::extract_swaps),
+                                _ => {
+                                    eprintln!("[Stream Task] Unknown protocol: {}", protocol_name);
+                                    return Vec::new();
                                 }
                             };
-                            match result {
+
+                            // Build filter for this protocol
+                            let mut filter = alloy::rpc::types::Filter::new()
+                                .from_block(batch_start)
+                                .to_block(batch_end)
+                                .event_signature(event_signature);
+                            if let Some(pool_addr) = pool_filter_local { filter = filter.address(pool_addr); }
+
+                            // Fetch logs via shard strategy or single call
+                            let logs_result = if service.config().shard_logs {
+                                match service.config().provider_strategy {
+                                    crate::config::ProviderStrategy::Race => {
+                                    service.rpc_service().get_logs_sharded(
+                                        event_signature,
+                                        pool_filter_local,
+                                        batch_start,
+                                        batch_end,
+                                        service.config().effective_shard_count(),
+                                    ).await
+                                    }
+                                    crate::config::ProviderStrategy::Shard => {
+                                        service.rpc_service().get_logs_provider_assigned_shards(
+                                            event_signature,
+                                            pool_filter_local,
+                                            batch_start,
+                                            batch_end,
+                                            service.config().effective_shard_count(),
+                                        ).await
+                                    }
+                                }
+                            } else {
+                                service.rpc_service().get_logs(&filter).await
+                            };
+
+                            match logs_result {
                                 Ok(raw_logs) => {
-                                    info!("✅ [{}] Got {} sharded logs for blocks {} to {}", protocol, raw_logs.len(), batch_start, batch_end);
+                                    info!("✅ [{}] Got {} logs for blocks {}-{}", protocol_name, raw_logs.len(), batch_start, batch_end);
                                     extract_fn(&raw_logs)
                                 }
                                 Err(e) => {
                                     let error_str = e.to_string();
                                     if error_str.contains("32701") || error_str.contains("Please specify an address") {
-                                        debug!("⚠️ [{}] Provider restriction - using fallback (expected behavior)", protocol);
+                                        debug!("⚠️ [{}] Provider restriction - expected fallback", protocol_name);
                                     } else {
-                                        warn!("❌ [{}] RPC error: {}", protocol, e);
+                                        warn!("❌ [{}] RPC error: {}", protocol_name, e);
                                     }
                                     Vec::new()
                                 }
                             }
-                        } else {
-                            match service.rpc_service().get_logs(&filter).await {
-                                Ok(raw_logs) => {
-                                    info!("✅ [{}] Got {} raw logs for blocks {} to {}", protocol, raw_logs.len(), batch_start, batch_end);
-                                    // Extract swaps using protocol-specific function
-                                    extract_fn(&raw_logs)
-                                }
-                                Err(e) => {
-                                    let error_str = e.to_string();
-                                    if error_str.contains("32701") || error_str.contains("Please specify an address") {
-                                        debug!("⚠️ [{}] Provider restriction - using fallback (expected behavior)", protocol);
-                                    } else {
-                                        warn!("❌ [{}] RPC error: {}", protocol, e);
-                                    }
-                                Vec::new()
-                                }
-                            }
-                        };
+                        }));
+                    }
 
-                        info!("🔄 [{}] Extracted {} swap events", protocol, logs.len());
-                        all_extracted_events.extend(logs);
+                    while let Some(logs) = proto_tasks.next().await {
+                        if !logs.is_empty() {
+                            info!("🔄 Extracted {} swap events from concurrent protocol task", logs.len());
+                            all_extracted_events.extend(logs);
+                        }
                     }
 
                     if !all_extracted_events.is_empty() {
@@ -883,6 +922,7 @@ pub fn create_batch_from_swap_events(swap_events: &[SwapEvent]) -> Result<Record
                 Arc::new(Int64Array::from(Vec::<Option<i64>>::new())), // block_timestamp
                 Arc::new(StringArray::from(Vec::<String>::new())), // tx_hash
                 Arc::new(UInt64Array::from(Vec::<u64>::new())), // log_index
+                Arc::new(UInt64Array::from(Vec::<Option<u64>>::new())), // tx_index
                 Arc::new(StringArray::from(Vec::<String>::new())), // dex_protocol
                 Arc::new(StringArray::from(Vec::<String>::new())), // pool_address
                 // Participants
@@ -904,7 +944,6 @@ pub fn create_batch_from_swap_events(swap_events: &[SwapEvent]) -> Result<Record
                 Arc::new(StringArray::from(Vec::<Option<String>>::new())), // token_sold_amount_raw
                 Arc::new(Float64Array::from(Vec::<Option<f64>>::new())),   // token_bought_amount
                 Arc::new(Float64Array::from(Vec::<Option<f64>>::new())),   // token_sold_amount
-                Arc::new(StringArray::from(Vec::<Option<String>>::new())), // trade_direction
                 // Transaction context
                 Arc::new(StringArray::from(Vec::<Option<String>>::new())), // tx_from
                 Arc::new(StringArray::from(Vec::<Option<String>>::new())), // tx_to
@@ -922,6 +961,7 @@ pub fn create_batch_from_swap_events(swap_events: &[SwapEvent]) -> Result<Record
     let mut tx_hashes = Vec::with_capacity(num_rows);
     let mut log_indices = Vec::with_capacity(num_rows);
     let mut dex_protocols = Vec::with_capacity(num_rows);
+    let mut tx_indices = Vec::with_capacity(num_rows);
     let mut pool_addresses = Vec::with_capacity(num_rows);
     let mut takers = Vec::with_capacity(num_rows);
     let mut recipients = Vec::with_capacity(num_rows);
@@ -943,7 +983,6 @@ pub fn create_batch_from_swap_events(swap_events: &[SwapEvent]) -> Result<Record
     let mut token_sold_amount_raws = Vec::with_capacity(num_rows);
     let mut token_bought_amounts = Vec::with_capacity(num_rows);
     let mut token_sold_amounts = Vec::with_capacity(num_rows);
-    let mut trade_directions = Vec::with_capacity(num_rows);
 
     // Transaction context arrays
     let mut tx_froms = Vec::with_capacity(num_rows);
@@ -962,6 +1001,7 @@ pub fn create_batch_from_swap_events(swap_events: &[SwapEvent]) -> Result<Record
         tx_hashes.push(event.tx_hash.clone());
         log_indices.push(event.log_index);
         dex_protocols.push(event.dex_protocol.clone());
+        tx_indices.push(event.tx_index);
         pool_addresses.push(event.pool_address.clone());
 
         // Participants
@@ -985,7 +1025,6 @@ pub fn create_batch_from_swap_events(swap_events: &[SwapEvent]) -> Result<Record
         token_sold_amount_raws.push(event.token_sold_amount_raw.clone());
         token_bought_amounts.push(event.token_bought_amount);
         token_sold_amounts.push(event.token_sold_amount);
-        trade_directions.push(event.trade_direction.clone());
 
         // Transaction context
         tx_froms.push(event.tx_from.clone());
@@ -1007,6 +1046,7 @@ pub fn create_batch_from_swap_events(swap_events: &[SwapEvent]) -> Result<Record
             Arc::new(Int64Array::from(block_timestamps)),
             Arc::new(StringArray::from(tx_hashes)),
             Arc::new(UInt64Array::from(log_indices)),
+            Arc::new(UInt64Array::from(tx_indices)),
             Arc::new(StringArray::from(dex_protocols)),
             Arc::new(StringArray::from(pool_addresses)),
             // Participants
@@ -1028,7 +1068,6 @@ pub fn create_batch_from_swap_events(swap_events: &[SwapEvent]) -> Result<Record
             Arc::new(StringArray::from(token_sold_amount_raws)),
             Arc::new(Float64Array::from(token_bought_amounts)),
             Arc::new(Float64Array::from(token_sold_amounts)),
-            Arc::new(StringArray::from(trade_directions)),
             // Transaction context
             Arc::new(StringArray::from(tx_froms)),
             Arc::new(StringArray::from(tx_tos)),
@@ -1038,10 +1077,6 @@ pub fn create_batch_from_swap_events(swap_events: &[SwapEvent]) -> Result<Record
             // USD analysis
             Arc::new(Float64Array::from(value_usds)),
             Arc::new(StringArray::from(value_usd_methods)),
-            // Enrichment extras
-            // (trade_direction is included earlier in the list to match schema order)
-            // Note: ensure vector order aligns with schema
-            // Already appended above as trade_direction
         ],
     )?;
 
@@ -1055,6 +1090,17 @@ pub fn apply_swap_filters(events: &[SwapEvent], config: &DextradesConfig) -> Vec
     events
         .iter()
         .filter(|event| {
+            // Router whitelist filter (requires transaction enricher)
+            if let Some(ref routers) = config.router_whitelist {
+                match &event.tx_to {
+                    Some(to) => {
+                        let pass = routers.iter().any(|r| r.eq_ignore_ascii_case(to));
+                        if !pass { return false; }
+                    }
+                    None => return false,
+                }
+            }
+
             // Volume filter - skip if token amounts are not available
             if let Some(_min_volume) = config.min_volume_filter {
                 // For now, we can only filter on token amounts since we don't have USD values
